function [model] = CV_LSCCA(Du, Dv, param, model, cv)

if strcmp(cv, 'sub2')
  % --------------------------------------------------
  % Search only subset in 2D grid points
  % --------------------------------------------------

  s_idx = 0;    
  l_idx = 0;

  cv_score = ones(length(param.sigma_list), length(param.lambda_list)) * Inf;
  lambda_start = 1;
  for sigma_cand = param.sigma_list
    s_idx = s_idx + 1;
    Ku = exp(- Du ./ (2*sigma_cand^2));    
    Kv = exp(- Dv ./ (2*sigma_cand^2));
    KuKv = Ku .* Kv;
    RegMat = model.coef_G*KuKv(:,param.ce_idx) + model.coef_I*eye(param.b);

    lambda_forward = [lambda_start:length(param.lambda_list)];
    lambda_back = [lambda_start-1:-1:1];

    lam_cnt = 0;
    while 1
      lam_cnt = lam_cnt + 1;
      
      if isempty(lambda_forward) % forward or backward
        forward_back = 2;
      else
        forward_back = 1;
      end
      
      if ~isempty(lambda_forward) % forward
        l_idx = lambda_forward(1);
        lambda_forward = lambda_forward(2:end);
        lambda_cand = param.lambda_list(l_idx);
      elseif ~isempty(lambda_back) % backward
        l_idx = lambda_back(1);
        lambda_back = lambda_back(2:end);
        lambda_cand = param.lambda_list(l_idx);
      else
        [min_cv lambda_start] = min(cv_score(s_idx, :));
        break;
      end
      
      for k = 1:param.fold_num
        tr_idx = param.cv_idx(param.cv_split ~= k);
        va_idx = param.cv_idx(param.cv_split == k);
        tr_ce_idx = ismember(param.ce_idx, tr_idx); % tr_ce_idx = ce_idx(tr_ce_idx);
        
        % -- Train --
        alpha_cv = uLSIF(Ku(tr_ce_idx,tr_idx), Kv(tr_ce_idx,tr_idx), ...
                         lambda_cand*(param.b/nnz(tr_ce_idx)), RegMat(tr_ce_idx,tr_ce_idx));
                         % lambda_cand, RegMat(tr_ce_idx,tr_ce_idx));

        % -- Validation --
        n_va = length(va_idx);
        Ku_va = Ku(tr_ce_idx,va_idx);
        Kv_va = Kv(tr_ce_idx,va_idx);
        Hhat_va = (Ku_va * Ku_va') .* (Kv_va * Kv_va') ./ (n_va^2);
        score_temp(k) = (alpha_cv' * Hhat_va * alpha_cv)/2 - ...
            alpha_cv' * mean(Ku_va .* Kv_va, 2);
        % (Ku_va .* Kv_va)' * alpha_cv
        % score_tmp(k) = alpha_cv' * mean(Ku_va .* Kv_va, 2);
      end

      cv_score(s_idx,l_idx) = mean(score_temp);

      if min(cv_score(s_idx,:)) < cv_score(s_idx,l_idx) && ...
            lam_cnt > 1 && s_idx > 1
        if  forward_back == 1 % forward
          lambda_forward = [];
        else
          lambda_back = [];
        end
      end
      
    end

  end

  % [param.sigma_list' cv_score]

  [min_score_row min_s_idx] = min(cv_score,[],1);
  [min_score min_l_idx] = min(min_score_row);
  model.sigmaX = param.sigma_list(min_s_idx(min_l_idx));
  model.sigmaY = model.sigmaX;
  model.lambda = param.lambda_list(min_l_idx);
  % sigmaY = d_V*sigma_list(min_s_idx(min_l_idx));
  
elseif strcmp(cv, 'ps3')

  % --------------------------------------------------
  % Pattern Search in 3 dim space
  % --------------------------------------------------
  check_table = zeros(length(param.lambda_list), length(param.sigma_list), length(param.sigma_list));

  Ku = exp(- Du ./ (2*model.sigmaX^2));    
  Kv = exp(- Dv ./ (2*model.sigmaY^2));
  KuKv = Ku .* Kv;
  RegMat = model.coef_G*KuKv(:,param.ce_idx) + model.coef_I*eye(param.b);

  for k = 1:param.fold_num
    tr_idx = param.cv_idx(param.cv_split ~= k);
    va_idx = param.cv_idx(param.cv_split == k);
    tr_ce_idx = ismember(param.ce_idx, tr_idx); % ce_idx(tr_ce_idx)
            
    %-- Train --
    alpha_cv = uLSIF(Ku(tr_ce_idx,tr_idx), Kv(tr_ce_idx,tr_idx), ...
                     model.lambda, RegMat(tr_ce_idx,tr_ce_idx));
            
    %-- Validation --
    n_va = length(va_idx);
    Ku_va = Ku(tr_ce_idx,va_idx);
    Kv_va = Kv(tr_ce_idx,va_idx);
    Hhat_va = (Ku_va * Ku_va') .* (Kv_va * Kv_va') ./ (n_va^2);        
    score_tmp(k) = (alpha_cv' * Hhat_va * alpha_cv)/2 - ...
        alpha_cv' * mean(Ku_va .* Kv_va, 2);
    
  end
  % fprintf('l = %f, sx = %f, sy = %f, cv = %f\n', ...
  % model.lambda, model.sigmaX, model.sigmaY, mean(score_tmp));
  min_score = mean(score_tmp);


  find_min = 1;
  ps_iter = 0;
  max_ps_iter = inf;
  while find_min == 1 && ps_iter < max_ps_iter
    ps_iter = ps_iter + 1;
    find_min = 0;

    [min_val sigmaX_idx] = min(abs(param.sigma_list - model.sigmaX));
    [min_val sigmaY_idx] = min(abs(param.sigma_list - model.sigmaY));
    [min_val lambda_idx] = min(abs(param.lambda_list - model.lambda));
    % sigmaX_idx = find(sigma_list == sigmaX);
    %sigmaY_idx = find(sigma_list == sigmaY);
    %lambda_idx = find(lambda_list == lambda);
    check_table(lambda_idx,sigmaX_idx,sigmaY_idx) = 1;

    % --
    pattern_mat = [eye(3); -eye(3)];

    sigmaX_idx_prev = sigmaX_idx;
    sigmaY_idx_prev = sigmaY_idx; 
    lambda_idx_prev = lambda_idx;
    for p = 1:size(pattern_mat,1)
      temp = [sigmaX_idx_prev sigmaY_idx_prev lambda_idx_prev] + pattern_mat(p,:);
      sigmaX_cand_idx = temp(1);
      sigmaY_cand_idx = temp(2);
      lambda_cand_idx = temp(3);

      if sigmaX_cand_idx > length(param.sigma_list) || sigmaX_cand_idx < 1 || ...
            sigmaY_cand_idx > length(param.sigma_list) || sigmaY_cand_idx < 1 || ...
            lambda_cand_idx > length(param.lambda_list) || lambda_cand_idx < 1 || ...
            check_table(lambda_cand_idx,sigmaX_cand_idx,sigmaY_cand_idx)
        continue;
      end
      sigmaX_cand = param.sigma_list(sigmaX_cand_idx);
      Ku = exp(- Du ./ (2*sigmaX_cand^2));
      sigmaY_cand = param.sigma_list(sigmaY_cand_idx);
      Kv = exp(- Dv ./ (2*sigmaY_cand^2));
      KuKv = Ku .* Kv;
      RegMat = model.coef_G*KuKv(:,param.ce_idx) + model.coef_I*eye(param.b);
      lambda_cand = param.lambda_list(lambda_cand_idx);
      
      for k = 1:param.fold_num
        tr_idx = param.cv_idx(param.cv_split ~= k);
        va_idx = param.cv_idx(param.cv_split == k);
        tr_ce_idx = ismember(param.ce_idx, tr_idx); % ce_idx(tr_ce_idx)
        
        %-- Train --
        alpha_cv = uLSIF(Ku(tr_ce_idx,tr_idx), Kv(tr_ce_idx,tr_idx), ...
                         lambda_cand*(param.b/nnz(tr_ce_idx)), RegMat(tr_ce_idx,tr_ce_idx));
                         % lambda_cand, RegMat(tr_ce_idx,tr_ce_idx));
                         % lambda_cand*(param.b/nnz(tr_ce_idx)), RegMat(tr_ce_idx,tr_ce_idx));
                         % lambda_cand, RegMat(tr_ce_idx,tr_ce_idx));
            
        %-- Validation --
        n_va = length(va_idx);
        Ku_va = Ku(tr_ce_idx,va_idx);
        Kv_va = Kv(tr_ce_idx,va_idx);
        Hhat_va = (Ku_va * Ku_va') .* (Kv_va * Kv_va') ./ (n_va^2);        
        score_tmp(k) = (alpha_cv' * Hhat_va * alpha_cv)/2 - ...
            alpha_cv' * mean(Ku_va .* Kv_va, 2);
        % (Ku_va .* Kv_va)' * alpha_cv
        % score_tmp(k) = alpha_cv' * mean(Ku_va .* Kv_va, 2);
      end
      % fprintf('l = %f, sx = %f, sy = %f, cv = %f\n', ...
      % lambda_cand, sigmaX_cand, sigmaY_cand, mean(score_tmp));
      check_table(lambda_cand_idx,sigmaX_cand_idx,sigmaY_cand_idx) = 1;
      if mean(score_tmp) < min_score
        min_score = mean(score_tmp);
        model.sigmaX = sigmaX_cand;
        model.sigmaY = sigmaY_cand;
        model.lambda = lambda_cand;
        find_min = 1;
      end
    end

  end % while 

end
