function [is_end U V param_optimizer model] = ...
    ng_update(obj_func, obj_value, X, Y, U, V, dU, dV, model, param, param_optimizer)

  is_end = 0;

  if param.verbose
    fprintf('update U V by natural gradient\n');
  end

  Delta_U = U'*dU - dU'*U;
  Delta_V = V'*dV - dV'*V;
  U_prev = U;
  V_prev = V;

  if strcmp(param_optimizer.line_search, 'naive')
    % --------------------------------------------------
    % Naive line search
    % --------------------------------------------------
    max_line_search = 20;
    %-- Normalize gradient --
    deno = (sqrt(trace(Delta_U*Delta_U')) ...
            + sqrt(trace(Delta_V*Delta_V')));
    if param.update_x && param.update_y
      deno = deno * 0.5;
    end
    Delta_U = Delta_U / deno;
    Delta_V = Delta_V / deno;

    step = 1;

    for i = 1:max_line_search
      U = U_prev * expm(param.update_x*step*Delta_U);
      V = V_prev * expm(param.update_y*step*Delta_V);
      if param.verbose
        fprintf('U = '); fprintf('%f ', U); fprintf('\n');
        fprintf('V = '); fprintf('%f ', V); fprintf('\n');
      end
      
      [obj model] = obj_func(U*X, V*Y, param, model, 'none');
      if param.verbose
        fprintf('obj = %f \n', obj);
      end
      
      step = step + 1;
      obj_log(i) = obj;
      if i > 1 && obj_log(i) < obj_log(i-1)
        break;
      end
    end

    [max_obj_temp idx] = max(obj_log);

    if max_obj_temp < obj_value
      U = U_prev;
      V = V_prev;
      obj = obj_value;
      if param.verbose
        fprintf('Naive line search failed\n');      
      end
      is_end = 1;
    else
      U = U_prev * expm(param.update_x*idx*Delta_U);
      V = V_prev * expm(param.update_y*idx*Delta_V);
    end

  else
    % --------------------------------------------------
    % Armijo's rule
    % --------------------------------------------------
    
    max_line_search = 10;      
    rho = 0.3; % Decay rate
    c = 0.1;   %0.0001

    step = 1;
    for i = 1:max_line_search
      U = U_prev * expm(param.update_x*step*Delta_U);
      V = V_prev * expm(param.update_y*step*Delta_V);

      if param.verbose
        fprintf('U = '); fprintf('%f ', U); fprintf('\n');
        fprintf('V = '); fprintf('%f ', V); fprintf('\n');
      end

      [obj model] = obj_func(U*X, V*Y, param, model, 'none');
      if param.verbose
        fprintf('obj = %f \n', obj);
      end

      armijo_rhs = - obj_value + step*c* ...  
          (trace(Delta_U'*U_prev'*(-dU)) + ... % geodesic * gradient
           trace(Delta_V'*V_prev'*(-dV)));      
      if -obj < armijo_rhs
        break;
      end
      step = rho * step;
    end

    if ((norm(Delta_U) <  0.00005) && (norm(Delta_V) <  0.00005)) || ...
          - obj_value + obj < param_optimizer.epsilon
      % (- obj_value + obj)/abs(obj) < param_optimizer.epsilon
      if param.verbose
        fprintf('Objective function does not inrease sufficiently.\n');
      end
      is_end = 1;
    end;
    
  end

