function [is_end U V param_optimizer model] = ...
      sqp_update(obj_func, obj_value, X, Y, U, V, dU, dV, model, param, param_optimizer)

  is_end = 0;

  if param_optimizer.iter > 1
    U_prev = param_optimizer.U_prev;
    V_prev = param_optimizer.V_prev;
    Q  = param_optimizer.Q;
    dlag_prev = param_optimizer.dlag_prev;
    lag_mlp = param_optimizer.lag_mlp;
  else
    param_optimizer.refresh_hessian = 1;
  end

  Sigma_X = eye(param.dim_X); % cov(X');
  Sigma_Y = eye(param.dim_Y); % cov(Y');
  dim_U = size(param.opt_U, 1) + 1;
  dim_V = size(param.opt_V, 1) + 1;

  if param.update_x && param.update_y
    grad = - [dU'; dV'];
  elseif ~param.update_x
    grad = - dV';
  elseif ~param.update_y
    grad = - dU';
  end

  if param_optimizer.iter > 1 && ~param_optimizer.refresh_hessian
    % --------------------------------------------------
    % BFGS update
    % --------------------------------------------------
    if param.update_x && param.update_y
      s = ([U V] - [U_prev V_prev])';
      C = [[param.opt_U; 2*U]*Sigma_X  zeros(dim_U, param.dim_Y); ...
           zeros(dim_V, param.dim_X)  [param.opt_V; 2*V]*Sigma_Y];
    elseif ~param.update_x
      s = (V - V_prev)';
      C = [param.opt_V; 2*V] * Sigma_Y;
    elseif ~param.update_y
      s = (U - U_prev)';
      C = [param.opt_U; 2*U] * Sigma_X;
    end
    dlag = grad + C'*lag_mlp;

    q = dlag - dlag_prev;
    sq = s' * q;

    Qs = Q * s;
    sQs = s' * Qs;

    if sq >= 0.2 * sQs
      qhat = q;
    else
      psi = (0.8*sQs) / (sQs - sq);
      qhat = psi*q + (1-psi)*Qs;
    end
    Q = Q - (Qs*Qs')/sQs + qhat*qhat'/sq;

    condQ = cond(Q);
    if condQ > 10^6 || sQs == 0 || sq == 0
      Q = eye(param.update_x*param.dim_X + param.update_y*param.dim_Y);
    end
        
    % aH * s - qhat

  else
    Q = eye(param.update_x*param.dim_X + param.update_y*param.dim_Y);
    param_optimizer.refresh_hessian = 0;
  end
  param_optimizer.Q = Q;

  % --------------------------------------------------
  % Solve QP
  % --------------------------------------------------
  Sigma_U = Sigma_X * U';
  Sigma_V = Sigma_Y * V';
  Utemp = [param.opt_U; U];
  Vtemp = [param.opt_V; V];
  USigma = Utemp * Sigma_X;
  VSigma = Vtemp * Sigma_Y;

  if param.update_x && param.update_y
    C = [[param.opt_U; 2*U]*Sigma_X  zeros(size(USigma,1),size(VSigma,2)); ...
         zeros(size(VSigma,1),size(USigma,2))  [param.opt_V; 2*V]*Sigma_Y];
    const_rhs = [USigma * U'; VSigma * V'] - ...
        [zeros(size(param.opt_U,1),1); 1; zeros(size(param.opt_V,1),1); 1];
  elseif ~param.update_x
    C = [param.opt_V; 2*V] * Sigma_Y;
    const_rhs = VSigma * V' - [zeros(size(param.opt_V,1),1); 1];
  elseif ~param.update_y
    C = [param.opt_U; 2*U] * Sigma_X
    const_rhs = USigma * U' - [zeros(size(param.opt_U,1),1); 1];
  end

  invQ = inv(Q);
  CinvQ = C * invQ;

  lastwarn('');
  lag_mlp = - (CinvQ*C') \ (CinvQ*grad - const_rhs);
  if strcmp(lastwarn,'Matrix is singular to working precision.') 
    fprintf('matrix close to singular');
    r = 10^(-6)*mean(mean(abs(CinvQ*C')))
    lag_mlp = - (CinvQ*C' + r*eye(size(C,1))) \ (CinvQ*grad - const_rhs);
  end

  % grad
  % C' * lag_mlp
  delta = - invQ * (grad + C' * lag_mlp);

  delta_U = delta(1:param.dim_X)';
  delta_V = delta(end-param.dim_Y+1:end)';
  param_optimizer.dlag_prev = grad + C' * lag_mlp;
  param_optimizer.lag_mlp = lag_mlp;

  % --------------------------------------------------
  % Line search
  % --------------------------------------------------
  
  rho = 0.3; % decay rate
  eta = 10^(-1); %0.1; % 10^(-4); %0.1; % Coef.
  initial_step = 1;
  max_line_search = 20;

  param_optimizer.U_prev = U;
  param_optimizer.V_prev = V;
  if param.update_x && param.update_y
    constraint = abs([Utemp * Sigma_X * U'; Vtemp * Sigma_Y * V'] - ...
                     [zeros(size(param.opt_U,1),1); 1; zeros(size(param.opt_V,1),1); 1]);
    c_prev = sum(abs([Utemp * Sigma_X * U'; Vtemp * Sigma_Y * V'] - ...
                     [zeros(size(param.opt_U,1),1); 1; zeros(size(param.opt_V,1),1); 1]));
  elseif ~param.update_x
    c_prev = sum(abs(Vtemp * Sigma_Y * V' - [zeros(size(param.opt_V,1),1); 1]));
  elseif ~param.update_y
    c_prev = sum(abs(Utemp * Sigma_X * U' - [zeros(size(param.opt_U,1),1); 1]));
  end

  % fprintf('sum(|c|) = %f\n', c_prev);

  if c_prev < 10^(-4)
    mu = 1.5 * max(abs(lag_mlp)); % + 0.1 * iter; %mu = 10 * max(abs(lag_mlp));
  else
    mu = max(max(abs(lag_mlp)), (grad' * delta) / ((1 - 0.1) * c_prev)) ...
         + 1.0;
  end
  if mu > 10^6
    mu = 10^6
  end

  obj_prev = obj_value;
  phi_prev = - obj_prev + mu * c_prev;

  if param.verbose
    fprintf('phi_prev = %f, obj = %f, mu*c = %f\n', ...
            phi_prev, obj_prev, mu*c_prev);
  end

  alpha = 1;
  for i = 1:max_line_search
    if param.update_x
      U = param_optimizer.U_prev + alpha * delta_U;
    end
    if param.update_y
      V = param_optimizer.V_prev + alpha * delta_V;
    end

    % fprintf('U = '); fprintf('%f ', U); fprintf('\n');
    % fprintf('V = '); fprintf('%f ', V); fprintf('\n');
    
    obj = obj_func(U*X, V*Y, param, model, 'none');% obj_func(U*X, V*Y, param, sigma, 0);
    % fprintf('obj = %f \n', obj);

    Utemp = [param.opt_U; U];
    Vtemp = [param.opt_V; V];

    if param.update_x && param.update_y
      c = sum(abs([Utemp * Sigma_X * U'; Vtemp * Sigma_Y * V'] - ...
                  [zeros(size(param.opt_U,1),1); 1; zeros(size(param.opt_V,1),1); 1]));
    elseif ~param.update_x
      c = sum(abs(Vtemp * Sigma_Y * V' - [zeros(size(param.opt_V,1),1); 1]));
    elseif ~param.update_y
      abs(Utemp * Sigma_X * U' - [zeros(size(param.opt_U,1),1); 1])
      c = sum(abs(Utemp * Sigma_X * U' - [zeros(size(param.opt_U,1),1); 1]))
    end

    phi = - obj + mu * c; % Merit function
    armijo = phi_prev + alpha*eta*(grad'*delta - mu*c);
    if param.verbose
      fprintf('obj = %f, mu*c = %f, phi = %f\n', obj, mu*c, phi);
    end

    if phi < phi_prev + alpha*eta*(grad'*delta - mu*c)
      break;
    end
    alpha = alpha * rho;
  end
  
  dif = [];
  if param.update_x
    dif = U - param_optimizer.U_prev;
  end
  if param.update_y
    dif = [dif V - param_optimizer.V_prev];
  end

  if abs(c) < 10^(-6)
    param_optimizer.feasible = 1;
  else
    param_optimizer.feasible = 0;
  end

  if norm(dif) < 10^(-8)  || ... 
        phi_prev - phi < param_optimizer.epsilon % 10^(-6)
    % Iteration may converge
    param_optimizer.refresh_hessian = 1;
    
    if abs(c) > 10^(-6)
      % Make orthnormal constraints satisfied 
      if size(param.opt_U,1) > 0
        U = U - (param.opt_U' * (param.opt_U * U'))';
      end
      U = U / norm(U);
      if size(param.opt_V,1) > 0
        V = V - (param.opt_V' * (param.opt_V * V'))';
      end
      V = V / norm(V);
      c_prev = 0;
      param_optimizer.feasible = 1;
    end
    is_end = 1;
  end


